-- This table is filled by crawlers
-- uri: The address of the timeseries to be indexed
-- description: Information filled by the crawler to describe the timeseries. This field is supposed to contain a user-friendly description.
-- metadata: Information filled by crawler. The data here is ideally a JSON document (object with string values) with properties describing the timeseries in detail,
-- but may be any text that can be used to construct a full text search index.
-- textsearch: generated field for the purpose of full text search
-- last_seen: Every time a record is inserted or updated (even when no modification occurs) the last_seen timestamp is updated.
-- The purpose of last_seen is to manage outdated entries and to remove them eventually.

CREATE OR REPLACE FUNCTION sherlock_normalize(str varchar) RETURNS varchar
AS $$
BEGIN
    RETURN lower(regexp_replace(regexp_replace(regexp_replace(str, '[^a-zA-Z0-9]+', ' ', 'g'),'([a-z])([A-Z])','\1 \2','g'),'([A-Z][A-Z0-9])([a-z])','\1 \2','g'));
END;
$$
LANGUAGE plpgsql IMMUTABLE RETURNS NULL ON NULL INPUT;


CREATE OR REPLACE FUNCTION sherlock_tsquery(str varchar) RETURNS tsquery
AS $$
BEGIN
    RETURN to_tsquery('english',lower(
    regexp_replace(trim(regexp_replace(regexp_replace(regexp_replace(regexp_replace(regexp_replace(regexp_replace(regexp_replace(regexp_replace(regexp_replace(
    trim(str),
    '[^a-zA-Z0-9 \-]+', ':', 'g'), -- Remove special characters, keep "-" for negation
    '(^| )-([^- ])','\1!\2','g'), -- Convert "-" to tsquery negation "!", but only if in front of a separated word
    '\!([^ ]+)','!(\1)','g'), -- Put negated words in brackets before camelcase and special character conversion
    '[-:]',' ','g'), -- Special characters into word separators
    '\( +','(','g'), -- Trim within groups
    ' +\)',')','g'), -- Trim within groups
    '([a-z])([A-Z])','\1 \2','g'),  -- Camelcase 1
    '([A-Z][A-Z0-9])([a-z])','\1 \2','g'),  -- Camelcase 2
    '([^ \(\)\!]+)','\1:*','g')), -- For each word it is enough to match the beginning of the word
    ' +',' & ','g') -- AND-connect
    ));
END;
$$
LANGUAGE plpgsql IMMUTABLE RETURNS NULL ON NULL INPUT;


CREATE TABLE IF NOT EXISTS sherlock_catalog (
uri varchar,
description varchar NOT NULL,
metadata varchar NOT NULL,
textsearch tsvector GENERATED ALWAYS AS (to_tsvector('english', sherlock_normalize(uri || ' ' || description || ' ' || metadata))) STORED,
last_seen timestamptz,
PRIMARY KEY (uri)
);

CREATE OR REPLACE FUNCTION sherlock_catalog_last_seen() RETURNS trigger AS $$
    BEGIN
        NEW.last_seen := now();
	RETURN NEW;
    END;
$$ LANGUAGE plpgsql;

CREATE INDEX IF NOT EXISTS sherlock_catalog_textsearch ON sherlock_catalog USING GIN (textsearch);

CREATE INDEX IF NOT EXISTS sherlock_catalog_last_seen ON sherlock_catalog (last_seen);

DO $$ BEGIN
    CREATE TRIGGER last_seen BEFORE INSERT OR UPDATE ON sherlock_catalog FOR EACH ROW EXECUTE PROCEDURE sherlock_catalog_last_seen();
EXCEPTION
    WHEN OTHERS THEN NULL;
END $$;

-- Example INSERT for adding new data to the catalog
-- INSERT INTO sherlock_catalog VALUES ('https://test.com/mydata', 'Amazing data', 'best data of all time') ON CONFLICT (uri) DO UPDATE SET description = EXCLUDED.description, metadata = EXCLUDED.metadata;

-- Example for deleting outdated entries
-- DELETE FROM sherlock_catalog WHERE last_seen < now() - '10 days'::interval;

-- Example for a search query
-- SELECT * FROM sherlock_catalog WHERE textsearch @@ sherlock_tsquery('amazed by the best data');

-- Search query with ranking
-- WITH query AS (SELECT sherlock_tsquery('amazed by the best data') AS query) SELECT sherlock_catalog.*, ts_rank(textsearch, query) AS rank FROM sherlock_catalog, query WHERE textsearch @@ query ORDER BY rank DESC;
